// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2012 CERN (www.cern.ch)
 * Author: Alessandro Rubini <rubini@gnudd.com>
 */

#include <linux/version.h>
#include <linux/slab.h>
#include <linux/ipmi/fru.h>

/**
 * Allocate memory
 * @size: memory size
 *
 * Return: a pointer to the new allocated buffer
 */
void *fru_alloc(size_t size)
{
	return kzalloc(size, GFP_KERNEL);
}

/**
 * Get a board info area type-length value
 * @header: FRU header
 * @nr: type-length number
 *
 * Return: a `n` type-length
 */
static struct fru_type_length *
__fru_get_board_tl(struct fru_common_header *header, int nr)
{
	struct fru_board_info_area *bia;
	struct fru_type_length *tl;

	bia = fru_get_board_area(header);
	tl = bia->tl;
	while (nr > 0 && !fru_is_eof(tl)) {
		tl = fru_next_tl(tl);
		nr--;
	}
	if (fru_is_eof(tl))
		return NULL;
	return tl;
}

/**
 * Get the string from a type-length
 * @header: FRU header
 * @nr: type-length number
 *
 * Return: a pointer to an allocated string containing the type-length value
 */
static char *__fru_alloc_get_tl(struct fru_common_header *header, int nr)
{
	struct fru_type_length *tl;
	char *res;

	tl = __fru_get_board_tl(header, nr);
	if (!tl)
		return NULL;

	res = fru_alloc(fru_strlen(tl) + 1);
	if (!res)
		return NULL;
	return fru_strcpy(res, tl);
}

/**
 * Get the manufacturer name
 * @header: FRU header
 *
 * Return: a pointer to the manufacturer string (user must free the memory)
 */
char *fru_get_board_manufacturer(struct fru_common_header *header)
{
	return __fru_alloc_get_tl(header, 0);
}
EXPORT_SYMBOL(fru_get_board_manufacturer);

/**
 * Get the preduct name
 * @header: FRU header
 *
 * Return: a pointer to the produtct string (user must free the memory)
 */
char *fru_get_product_name(struct fru_common_header *header)
{
	return __fru_alloc_get_tl(header, 1);
}
EXPORT_SYMBOL(fru_get_product_name);

/**
 * Get the serial number
 * @header: FRU header
 *
 * Return: a pointer to the serial number string (user must free the memory)
 */
char *fru_get_serial_number(struct fru_common_header *header)
{
	return __fru_alloc_get_tl(header, 2);
}
EXPORT_SYMBOL(fru_get_serial_number);

/**
 * Get the part number
 * @header: FRU header
 *
 * Return: a pointer to the part number string (user must free the memory)
 */
char *fru_get_part_number(struct fru_common_header *header)
{
	return __fru_alloc_get_tl(header, 3);
}
EXPORT_SYMBOL(fru_get_part_number);
